%Ian Gemp
%The polysim function simulates a polygon bouncing on a moving table for a
%certain period of time or for a certain number of polygon/table contacts.

%Basic Outline
%The function checks if the polygon is below the table after each time
%step. If the polygon is below the table after a time step, the function
%finds the exact polygon/table contact time and state.  The function then
%calculates the post impact state of the polygon and repeats the above
%steps.

%The input parameters to the polysim function are explained below.
%The output parameters to the polysim function are the frames which contain
%the picture-by-picture information to playback a movie of the simulation,
%the contact times and states of the polygon at each impact, and a list of
%the time steps used by the funciton.

%Note: Saving Data
%use save 'filename' variable -ascii
%use load 'filename'


function [F,contstates,object,table,steplist] = polysim(object,state,table,params,option)

close all
%note: options: 1)playnow & store movie 2)contact states only - Tlim
%3)contact states only - hitlim

%{
object.nvert = number of vertices
object.x = x location of each vertex in relation to center of mass
object.y = y location of each vertex in relation to center of mass
object.r = distance from each vertex to center of mass
object.theta = original angle of each vertex in relation to center of mass
object.inertia = inertia of polygon
object.u = friction coefficient
object.res = restitution coefficient

state.x = x position of polygon center of mass
state.y = y position of polygon center of mass
state.theta = rotation angle of polygon about center of mass
state.xd = x velocity of polygon center of mass
state.yd = y velocity of polygon center of mass
state.thetad = rotation velocity of polygon about center of mass

%state(2) stores prediction for next state
state(2).x = x position of polygon center of mass
state(2).y = y position of polygon center of mass
state(2).theta = rotation angle of polygon about center of mass
state(2).xd = x velocity of polygon center of mass
state(2).yd = y velocity of polygon center of mass
state(2).thetad = rotation velocity of polygon about center of mass

table.x = x function of table center
table.xd = dx/dt function of table center
table.y = y function of table center
table.yd = dy/dt function of table center
table.phi = angle function of table center
table.phid = dphi/dt function of table center
table.cloudmax = maximum table height (amplitude of sin/cos function)

params.step = simulation stepsize
params.stepmov = movie stepsize
params.T = time to run simulation to
params.hitlim = number of table/object contacts to run simulation to
params.windim = window plot dimensions
%}

%length of table that gets displayed in each simulation frame
params.L = params.windim(2);

%set up object.r and object.theta for polygon or ball
if object.nvert>1
    %set up object.r
    object.r = sqrt(object.x.^2+object.y.^2);
    %set up object.theta
    for i=1:object.nvert
        object.theta(i) = atan2(object.y(i),object.x(i));
    end
else
    %default angle = 0 for point
    object.theta = 0;
end

%set table functions (user gives x,y,theta as strings)
%user can also manually set xd,yd, and thetad velocities
%{
%still figuring this out
if exist('table.xd')
    table.xd = inline(table.xd);
else
    table.xd = inline(char(diff(table.x)));
end
if exist('table.yd')
    table.yd = inline(table.yd);
    'got here'
else
    table.yd = inline(char(diff(table.y)));
end
if exist('table.phid')
    table.phid = inline(table.phid);
else
    table.phid = inline(char(diff(table.phi)));
end
%}
%
table.xd = inline(char(diff(table.x)));
%table.yd = inline(table.yd);
table.yd = inline(char(diff(table.y)));
table.phid = inline(char(diff(table.phi)));
table.x = inline(table.x);
table.y = inline(table.y);
table.phi = inline(table.phi);

%gravitational acceleration
g = -9.81;

%initial energy
nrg0 = -state(1).y*g+1/2*state(1).yd^2 + 1/2*state(1).xd^2 + 1/2*object.inertia*state(1).thetad^2;

%contact states - 1st col = times, 2nd col = states
contstates = cell(1,2);
contstates{1} = 0;
contstates{2} = state;
%movie states
movstates = cell(1);
movstates{1} = state;

t = 0; %simulation time
j = 1; %counter for plotting
hit = 1; %counter for contact instances - 1 corresponds to initial object state

F = []; %holds simulation frames
if option==1 %if want to watch movie during live simulation
    F = polyplot(table,params,movstates{j},object,nrg0,t);
    donePlotting = 0;
    params.hitlim = 0;
    tic %plotting has begun
end

if option==2 %run simulation to a certain time
    params.hitlim = 0;
end

if option==3 %run simulation to a certain impact number
    params.T = 0;
end

%the zerofinder finds a time between an above "time" and a "below" time and
%dt is the impact time minus the "above" time
dt = 0;
step = params.step; %time step used by simulator (varies)
initstep = params.step; %initial (minimum) time step set by user
steplist = params.step; %initalized the list of time steps

%finding contact states and possibly ploting
while (t<params.T || (hit-1)<=params.hitlim)
    
    %Compute variable time step so that polygon is near table, but still
    %above it
    %calculate y distance from ball to table
    eps = .99; %for ball;
    if object.nvert>1
        eps = .8; %for polygon
    end
    a = 1/2*g;
    b = state(1).yd;
    ytab = table.cloudmax;
    c = (state(1).y-max(object.r)) - ytab; %max(object.r) is problem for irregular polygon
    tstep = real([eps*(-b-sqrt(b^2-4*a*c))/(2*a),eps*(-b+sqrt(b^2-4*a*c))/(2*a),initstep]);
    ind = tstep>=0;
    step = max(tstep(ind));
    %update steplist
    steplist(end+1) = step;
    
    %predict next obj position
    t = t + step;
    state(2).x = state(1).x + state(1).xd*(step);
    state(2).y = state(1).y + state(1).yd*(step) + 1/2*g*(step)^2;
    state(2).theta = state(1).theta + state(1).thetad*(step);
    state(2).xd = state(1).xd;
    state(2).yd = state(1).yd + g*(step);
    state(2).thetad = state(1).thetad;

    %check if below table
    below = 0;
    contpt = [];
    for i=1:object.nvert
        %point: (m,n)
        %line: Ax + By + C
        %distance from point to line = (Am + Bn + C)/sqrt(A^2+B^2)
        %negative distance corresponds to below line
        A = tan(table.phi(t));
        B = -1;
        C = table.y(t) - A*table.x(t);
        if object.nvert>1
            m = object.r(i)*cos(state(2).theta + object.theta(i)) + state(2).x;
            n = object.r(i)*sin(state(2).theta + object.theta(i)) + state(2).y;
            distance = -(A*m + B*n + C)/sqrt(A^2 + B^2);
        else
            m = state(2).x;
            n = state(2).y;
            distance = -(A*m + B*n + C)/sqrt(A^2 + B^2) - object.r;
        end

        if distance<=0 %if distance==0 don't use zerofinder
            below = 1;
            contpt = [contpt i];
        end
    end

    %find contact vertex and contact time and simulate impact
    if (below==1)
        if distance==0
            tmin = t;
        else
            'zero finding';
            hit = hit+ 1;
            tcontact = zeros(1,length(contpt));
            t0 = t-step;
            x0 = state(1).x;
            y0 = state(1).y;
            theta0 = state(1).theta;
            v0x = state(1).xd;
            v0y = state(1).yd;
            w0 = state(1).thetad;
            t1 = t;
            tolmax = 1e-8;
            amp = table.cloudmax*2;
            tolopt = [tolmax,amp*tolmax];
            tol = min(tolopt(tolopt>0));
            dttolopt = 1e-6*[tolmax,amp*tolmax];
            dttol = min(dttolopt(dttolopt>0));
            if object.nvert>1 %polygon
                for i=1:length(contpt)
                    objtheta = object.theta(contpt(i));
                    tcontact(i) = zerofinder(t0,x0,y0,theta0,v0x,v0y,w0,objtheta,t1,tol,table.y,table.x,table.phi,object.r(contpt(i)),object.nvert,dttol);
                end
            else %ball
                objtheta = object.theta;
                tcontact = zerofinder(t0,x0,y0,theta0,v0x,v0y,w0,objtheta,t1,tol,table.y,table.x,table.phi,object.r(i),object.nvert,dttol);
            end
            tmin = min(tcontact);
            %safeguard against a zero time step
            if ((tmin-t0)<dttol) %dt = tcontact-t0)
                'dt = 0: exiting simulation'
                return%exit program if dt~0
            end
        end

        %check if there are multiple points of contact
        %nothing implemented right now if there are multiple contact points
        contptreal = [];
        for z=1:length(tcontact)
            if tcontact(z)==tmin
                contptreal = [contptreal contpt(z)];
            end
        end
        if length(contptreal)>1
            disp('multiple points of contact');
        end

        %state at contact
        dt = tmin-t0;
        state(1).x = state(1).x + state(1).xd*dt;
        state(1).y = state(1).y + state(1).yd*dt + 1/2*g*dt^2;
        state(1).theta = state(1).theta + state(1).thetad*dt;
        state(1).xd = state(1).xd;
        state(1).yd = state(1).yd + g*dt;
        state(1).thetad = state(1).thetad;

        
        %object - compute normal and tangential state of ball with
        %relation to table
        phi = table.phi(tmin); %table angle
        if object.nvert>1
            xc = object.r(contptreal(1))*cos(state(1).theta + object.theta(contptreal(1))) + state(1).x;
            yc = object.r(contptreal(1))*sin(state(1).theta + object.theta(contptreal(1))) + state(1).y;

            %construct line that describes table
            m = state(1).x;
            n = state(1).y;
            A = tan(table.phi(tmin));
            B = -1;
            C = table.y(tmin) - A*table.x(tmin);
            ydistance = -(A*m + B*n + C)/sqrt(A^2 + B^2);

            m = state(1).y;
            n = state(1).x;
            A = tan(table.phi(tmin));
            B = -1;
            C = xc - A*yc;
            xdistance = -(A*m + B*n + C)/sqrt(A^2 + B^2);
        else
            xdistance = 0;
            ydistance = object.r(contptreal(1));
        end

        vx1o = state(1).yd*sin(phi) + state(1).xd*cos(phi);
        vy1o = state(1).yd*cos(phi) - state(1).xd*sin(phi);
        w1o = state(1).thetad;
        x1 = xdistance;
        y1 = ydistance;
        r1 = object.r(contptreal(1));

        vx2o = table.xd(tmin)*cos(phi)+table.yd(tmin)*sin(phi);
        vy2o = table.yd(tmin)*cos(phi)-table.xd(tmin)*sin(phi);
        w2o = 0;

        %object rest prevention
        %set coefficient of restitution to 1 if the ball is barely bouncing
        %on the table
        cores = object.res;
        %{
        tol = 7.5e-1;
        if abs((vy1o-vy2o)*amp)<tol
            cores = 1;
            'restitution set to 1'
        end
        %}

        %use impact model here: need everything in normal/tangential
        %coordinates
        [vx1,vy1,w1] = impact(vx1o,vy1o,w1o,x1,y1,r1,vx2o,vy2o,w2o,cores,object.u,object.inertia);

        %set post impact state of polygon
        state(1).xd = vx1;
        state(1).yd = vy1;
        state(1).thetad = w1;

        contstates{1}(hit) = tmin;
        contstates{2}(hit) = state(1);
        %interpolate movie states if necessary
        if option==1
            if (contstates{1}(hit)-contstates{1}(hit-1))>params.stepmov
                newstates = datarecon(contstates{1}(hit-1),contstates{1}(hit),contstates{2}(hit-1),params.stepmov);
                for ii=1:length(newstates)
                    movstates{end+1} = newstates{ii};
                end
            end
        end
        t = tmin;
    else
        %dt = 0; %look into this
        state(1) = state(2);
    end
    %simultaneous plotting
    if (option==1 && toc>j*params.stepmov && j+1<=length(movstates))
        F(j+1) = polyplot(table,params,movstates{j+1},object,nrg0,toc);
        j = j + 1;
    end

end

if option>1
    donePlotting=2;
elseif length(F)>=length(movstates)
    donePlotting=2;
end
numframes = length(movstates);
while donePlotting<1
    %simultaneous plotting
    if option==1 && toc>j*params.stepmov && j+1<=numframes
        F(j+1) = polyplot(table,params,movstates{j+1},object,nrg0,toc);
        j = j + 1;
    end
    if j>=numframes
        donePlotting=2;
    end
end

close all

%{
common initial conditions

%ball bouncing on flat table with res = 0.8
object.nvert = 1;
object.x = 0;
object.y = 0;
object.r = 1;
object.theta = 0;
object.inertia = 1;
object.u = 0;
object.res = .8;
state.x = 5;
state.xd = 0;
state.y = 5;
state.yd = 0;
state.theta = 0;
state.thetad = 0;
table.phi = '0';
table.x = '5';
table.y = '2';
table.cloudmax = 2;
params.step = .01;
params.stepmov = .2;
params.T = 5;
params.hitlim = 10;
params.windim = [0 10 0 10];
[F,contstates] = polysim(object,state,table,params,3);
F = movconstr(contstates,table,params,object);

%table moves sinusoidally with ball
object.nvert = 1;
object.x = 0;
object.y = 0;
object.r = 1;
object.theta = 0;
object.inertia = 1;
object.u = 0;
object.res = .8;
state.x = 5;
state.xd = 0;
state.y = 5;
state.yd = 0;
state.theta = 0;
state.thetad = 0;
table.phi = '0';
table.x = '5';
table.y = '0.5*sin(10*t)+2';
table.cloudmax = 2.5;
params.step = .01;
params.stepmov = .2;
params.T = 20;
params.hitlim = 100;
params.windim = [0 10 0 10];
[F,contstates] = polysim(object,state,table,params,2);
F = movconstr(contstates,table,params,object);

%polygon on stationary table
object.nvert = 6;
object.x = [1,2,1,-1.5,-1.75,-1];
object.y = [-.5,0,1.5,1.25,1,0];
object.inertia = 1;
object.u = 0;
object.res = 1;
state.x = 5;
state.xd = 0;
state.y = 7;
state.yd = 0;
state.theta = 0;
state.thetad = 0;
table.phi = '0';
table.x = '5';
table.y = '2';
table.cloudmax = 2;
params.step = .01;
params.stepmov = .2;
params.T = 20;
params.windim = [0 10 0 10];
[F,contstates,object] = polysim(object,state,table,params,2);
F = movconstr(contstates,table,params,object);

%square on stationary table
object.nvert = 4;
object.x = [1,-1,-1,1];
object.y = [1,1,-1,-1];
object.inertia = 1;
object.u = 0;
object.res = 1;
state.x = 5;
state.xd = 0;
state.y = 7;
state.yd = 0;
state.theta = pi/4.2;
state.thetad = 0;
table.phi = '0';
table.x = '5';
table.y = '2';
table.cloudmax = 2;
params.step = .01;
params.stepmov = .2;
params.T = 20;
params.windim = [0 10 0 10];
[F,contstates] = polysim(object,state,table,params,1);

%}